"""Extensions to scipy.stats functions."""
import numpy as np
import scipy.stats
from numpy.linalg import inv
from scipy.special import erf, logsumexp
from scipy.stats._multivariate import multivariate_normal_frozen

from lsbi.utils import bisect


class multivariate_normal(multivariate_normal_frozen):  # noqa: D101
    def marginalise(self, indices):
        """Marginalise over indices.

        Parameters
        ----------
        indices : array_like
            Indices to marginalise.
        """
        i = self._bar(indices)
        mean = self.mean[i]
        cov = self.cov[i][:, i]
        return multivariate_normal(mean, cov)

    def condition(self, indices, values):
        """Condition on indices with values.

        Parameters
        ----------
        indices : array_like
            Indices to condition over.
        values : array_like
            Values to condition on.
        """
        i = self._bar(indices)
        k = indices
        mean = self.mean[i] + self.cov[i][:, k] @ inv(self.cov[k][:, k]) @ (
            values - self.mean[k]
        )
        cov = (
            self.cov[i][:, i]
            - self.cov[i][:, k] @ inv(self.cov[k][:, k]) @ self.cov[k][:, i]
        )
        return multivariate_normal(mean, cov)

    def _bar(self, indices):
        """Return the indices not in the given indices."""
        k = np.ones(len(self.mean), dtype=bool)
        k[indices] = False
        return k

    def bijector(self, x, inverse=False):
        """Bijector between U([0, 1])^d and the distribution.

        - x in [0, 1]^d is the hypercube space.
        - theta in R^d is the physical space.

        Computes the transformation from x to theta or theta to x depending on
        the value of inverse.

        Parameters
        ----------
        x : array_like, shape (..., d)
            if inverse: x is theta
            else: x is x
        inverse : bool, optional, default=False
            If True: compute the inverse transformation from physical to
            hypercube space.
        """
        L = np.linalg.cholesky(self.cov)
        if inverse:
            Linv = inv(L)
            y = np.einsum("ij,...j->...i", Linv, x - self.mean)
            return scipy.stats.norm.cdf(y)
        else:
            y = scipy.stats.norm.ppf(x)
            return self.mean + np.einsum("ij,...j->...i", L, y)


class mixture_multivariate_normal(object):
    """Mixture of multivariate normal distributions.

    Implemented with the same style as scipy.stats.multivariate_normal

    Parameters
    ----------
    means : array_like, shape (n_components, n_features)
        Mean of each component.

    covs: array_like, shape (n_components, n_features, n_features)
        Covariance matrix of each component.

    logA: array_like, shape (n_components,)
        Log of the mixing weights.
    """

    def __init__(self, means, covs, logA):
        self.means = np.array([np.atleast_1d(m) for m in means])
        self.covs = np.array([np.atleast_2d(c) for c in covs])
        self.logA = np.atleast_1d(logA)

    def logpdf(self, x, reduce=True, keepdims=False):
        """Log of the probability density function."""
        process_quantiles = scipy.stats.multivariate_normal._process_quantiles
        x = process_quantiles(x, self.means.shape[-1])
        dx = self.means - x[..., None, :]
        invcovs = np.linalg.inv(self.covs)
        chi2 = np.einsum("...ij,ijk,...ik->...i", dx, invcovs, dx)
        norm = -np.linalg.slogdet(2 * np.pi * self.covs)[1] / 2
        logpdf = norm - chi2 / 2
        if reduce:
            logA = self.logA - scipy.special.logsumexp(self.logA)
            logpdf = np.squeeze(scipy.special.logsumexp(logpdf + logA, axis=-1))
        if not keepdims:
            logpdf = np.squeeze(logpdf)
        return logpdf

    def rvs(self, size=1):
        """Random variates."""
        size = np.atleast_1d(size)
        p = np.exp(self.logA - self.logA.max())
        p /= p.sum()
        i = np.random.choice(len(p), size, p=p)
        x = np.random.randn(*size, self.means.shape[-1])
        choleskys = np.linalg.cholesky(self.covs)
        return np.squeeze(self.means[i, ..., None] + choleskys[i] @ x[..., None])

    def marginalise(self, indices):
        """Marginalise over indices.

        Parameters
        ----------
        indices : array_like
            Indices to marginalise.
        """
        i = self._bar(indices)
        means = self.means[:, i]
        covs = self.covs[:, i][:, :, i]
        logA = self.logA
        return mixture_multivariate_normal(means, covs, logA)

    def condition(self, indices, values):
        """Condition on indices with values.

        Parameters
        ----------
        indices : array_like
            Indices to condition over.
        values : array_like
            Values to condition on.
        """
        i = self._bar(indices)
        k = indices
        marginal = self.marginalise(i)

        means = self.means[:, i] + np.einsum(
            "ija,iab,ib->ij",
            self.covs[:, i][:, :, k],
            inv(self.covs[:, k][:, :, k]),
            (values - self.means[:, k]),
        )
        covs = self.covs[:, i][:, :, i] - np.einsum(
            "ija,iab,ibk->ijk",
            self.covs[:, i][:, :, k],
            inv(self.covs[:, k][:, :, k]),
            self.covs[:, k][:, :, i],
        )
        logA = (
            marginal.logpdf(values, reduce=False) + self.logA - marginal.logpdf(values)
        )
        return mixture_multivariate_normal(means, covs, logA)

    def _bar(self, indices):
        """Return the indices not in the given indices."""
        k = np.ones(self.means.shape[-1], dtype=bool)
        k[indices] = False
        return k

    def bijector(self, x, inverse=False):
        """Bijector between U([0, 1])^d and the distribution.

        - x in [0, 1]^d is the hypercube space.
        - theta in R^d is the physical space.

        Computes the transformation from x to theta or theta to x depending on
        the value of inverse.

        Parameters
        ----------
        x : array_like, shape (..., d)
            if inverse: x is theta
            else: x is x
        inverse : bool, optional, default=False
            If True: compute the inverse transformation from physical to
            hypercube space.
        """
        theta = np.empty_like(x)
        if inverse:
            theta[:] = x
            x = np.empty_like(x)

        for i in range(x.shape[-1]):
            m = self.means[..., :, i] + np.einsum(
                "ia,iab,...ib->...i",
                self.covs[:, i, :i],
                inv(self.covs[:, :i, :i]),
                theta[..., None, :i] - self.means[:, :i],
            )
            c = self.covs[:, i, i] - np.einsum(
                "ia,iab,ib->i",
                self.covs[:, i, :i],
                inv(self.covs[:, :i, :i]),
                self.covs[:, i, :i],
            )
            dist = mixture_multivariate_normal(
                self.means[:, :i], self.covs[:, :i, :i], self.logA
            )
            logA = (
                self.logA
                + dist.logpdf(theta[..., :i], reduce=False, keepdims=True)
                - dist.logpdf(theta[..., :i], keepdims=True)[..., None]
            )
            A = np.exp(logA - logsumexp(logA, axis=-1)[..., None])

            def f(t):
                return (A * 0.5 * (1 + erf((t[..., None] - m) / np.sqrt(2 * c)))).sum(
                    axis=-1
                ) - y

            if inverse:
                y = 0
                x[..., i] = f(theta[..., i])
            else:
                y = x[..., i]
                a = (m - 10 * np.sqrt(c)).min(axis=-1)
                b = (m + 10 * np.sqrt(c)).max(axis=-1)
                theta[..., i] = bisect(f, a, b)
        if inverse:
            return x
        else:
            return theta



import pytest
from numpy.testing import assert_allclose

from lsbi.utils import bisect


def test_bisect():
    def f(x):
        return x - 5

    assert bisect(f, 0, 10) == 5

    with pytest.raises(ValueError):
        bisect(f, 0, 4)

    def f(x):
        return x - [1, 2]

    assert_allclose(bisect(f, 0, 10), [1, 2])



"""Simple binary classifiers to perform model comparison."""
import torch
import torch.nn as nn
import torch.optim as optim
from torch.optim.lr_scheduler import ExponentialLR


class BinaryClassifierBase(nn.Module):
    """Base model for binary classification. Following 2305.11241.

    A simple binary classifier:
        - 5 hidden layers:
            - Layer 1 with initial_dim units
            - Layers 2-4 with internal_dim units
        - Leaky ReLU activation function
        - Batch normalization
        - Output layer with 1 unit linear classifier unit
        - Adam optimizer with default learning rate 0.001
        - Exponential learning rate decay with default decay rate 0.95

    Parameters
    ----------
    input_dim : int
        Dimension of the input data.
    internal_dim : int, optional (default=16)
        Dimension of the internal layers of the network.
    initial_dim : int, optional (default=130)
        Dimension of the first layer of the network.
    """

    def __init__(self, input_dim, internal_dim=16, initial_dim=130):
        super(BinaryClassifierBase, self).__init__()

        self.model = nn.Sequential(
            nn.Linear(input_dim, initial_dim),
            nn.LeakyReLU(),
            nn.BatchNorm1d(initial_dim),
            nn.Linear(initial_dim, internal_dim),
            nn.LeakyReLU(),
            nn.BatchNorm1d(internal_dim),
            nn.Linear(internal_dim, internal_dim),
            nn.LeakyReLU(),
            nn.BatchNorm1d(internal_dim),
            nn.Linear(internal_dim, internal_dim),
            nn.LeakyReLU(),
            nn.BatchNorm1d(internal_dim),
            nn.Linear(internal_dim, internal_dim),
            nn.LeakyReLU(),
            nn.Linear(internal_dim, 1),
        )

    def forward(self, x):
        """Forward pass through the network, logit output."""
        return self.model(x)

    def loss(self, x, y):
        """Loss function for the network."""
        raise NotImplementedError

    def predict(self, x):
        """Predict the Bayes Factor."""
        raise NotImplementedError

    def fit(self, X, y, **kwargs):
        """Fit classifier on input features X to predict labels y.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Input data.
        y : array-like, shape (n_samples,)
            Target values.
        num_epochs : int, optional (default=10)
            Number of epochs to train the network.
        batch_size : int, optional (default=128)
            Batch size for training.
        decay_rate : float, optional (default=0.95)
            Decay rate for the learning rate scheduler.
        lr : float, optional (default=0.001)
            Learning rate for the optimizer.
        device : str, optional (default="cpu")
            Device to use for training.
        """
        num_epochs = kwargs.get("num_epochs", 10)
        batch_size = kwargs.get("batch_size", 128)
        decay_rate = kwargs.get("decay_rate", 0.95)
        lr = kwargs.get("lr", 0.001)
        device = torch.device(kwargs.get("device", "cpu"))

        print("Using device: ", device)

        # Convert labels to torch tensor
        X = torch.tensor(X, dtype=torch.float32)
        labels = torch.tensor(y, dtype=torch.float32)
        labels = labels.unsqueeze(1)
        labels = labels.to(device)

        # Create a DataLoader for batch training
        dataset = torch.utils.data.TensorDataset(X, labels)
        dataloader = torch.utils.data.DataLoader(
            dataset, batch_size=batch_size, shuffle=True
        )

        # Define the loss function and optimizer
        criterion = self.loss
        optimizer = optim.Adam(self.parameters(), lr=lr)

        # Create the scheduler and pass in the optimizer and decay rate
        scheduler = ExponentialLR(optimizer, gamma=decay_rate)

        # Create a DataLoader for batch training
        self.to(device=device, dtype=torch.float32)

        for epoch in range(num_epochs):
            epoch_loss = []
            for i, (inputs, targets) in enumerate(dataloader):
                # Clear gradients
                optimizer.zero_grad()
                inputs = inputs.to(device)
                # Forward pass
                loss = criterion(inputs, targets)
                epoch_loss.append(loss.item())
                # Backward pass and optimize
                loss.backward()
                optimizer.step()

            # Print loss for every epoch
            scheduler.step()
            mean_loss = torch.mean(torch.tensor(epoch_loss)).item()
            print(f"Epoch {epoch+1}/{num_epochs}, Loss: {mean_loss}")

        # once training is done, set the model to eval(), ensures batchnorm
        # and dropout are not used during inference
        self.model.eval()


class BinaryClassifier(BinaryClassifierBase):
    """
    Extends the BinaryClassifierBase to use a BCE loss function.

    Furnishes with a direction prediction of the Bayes Factor.
    """

    def loss(self, x, y):
        """Binary cross entropy loss function for the network."""
        y_ = self.forward(x)
        return nn.BCEWithLogitsLoss()(y_, y)

    def predict(self, x):
        """Predict the log Bayes Factor.

        log K = lnP(Class 1) - lnP(Class 0)
        """
        x = torch.tensor(x, dtype=torch.float32)
        pred = self.forward(x)
        pred = nn.Sigmoid()(pred)
        return (torch.log(pred) - torch.log(1 - pred)).detach().numpy()


class BinaryClassifierLPop(BinaryClassifierBase):
    """
    Extends the BinaryClassifierBase to use a LPop Exponential loss.

    Furnishes with a direction prediction of the Bayes Factor.

    Parameters
    ----------
    alpha : float, optional (default=2.0)
        Scale factor for the exponent transform.
    """

    def __init__(self, *args, **kwargs):
        self.alpha = kwargs.pop("alpha", 2.0)
        super(BinaryClassifierLPop, self).__init__(*args, **kwargs)

    def lpop(self, x):
        """Leaky parity odd power transform."""
        return x + x * torch.pow(torch.abs(x), self.alpha - 1.0)

    def loss(self, x, y):
        """Lpop Loss function for the network."""
        x = self.forward(x)
        return torch.exp(
            torch.logsumexp((0.5 - y) * self.lpop(x), dim=0)
            - torch.log(torch.tensor(x.shape[0], dtype=torch.float64))
        ).squeeze()

    def predict(self, x):
        """Predict the log Bayes Factor.

        log K = lnP(Class 1) - lnP(Class 0)
        """
        x = torch.tensor(x, dtype=torch.float32)
        pred = self.forward(x)
        pred = self.lpop(pred)
        return pred.detach().numpy()



"""Gaussian models for linear Bayesian inference."""
import numpy as np
from numpy.linalg import inv, solve

from lsbi.stats import mixture_multivariate_normal, multivariate_normal
from lsbi.utils import logdet


class LinearModel(object):
    """A linear model.

    D|theta ~ N( m + M theta, C )
    theta   ~ N( mu, Sigma )

    Defined by:
        Parameters:       theta (n,)
        Data:             D     (d,)
        Prior mean:       mu    (n,)
        Prior covariance: Sigma (n, n)
        Data mean:        m     (d,)
        Data covariance:  C     (d, d)

    Parameters
    ----------
    M : array_like, optional
        if matrix: model matrix
        if vector: diagonal matrix with vector on diagonal
        if scalar: scalar * rectangular identity matrix
        Defaults to rectangular identity matrix
    m : array_like, optional
        if vector: data mean
        if scalar: scalar * unit vector
        Defaults to zero vector
    C : array_like, optional
        if matrix: data covariance
        if vector: diagonal matrix with vector on diagonal
        if scalar: scalar * identity matrix
        Defaults to identity matrix
    mu : array_like, optional
        if vector: prior mean
        if scalar: scalar * unit vector
        Defaults to zero vector
    Sigma : array_like, optional
        if matrix: prior covariance
        if vector: diagonal matrix with vector on diagonal
        if scalar: scalar * identity matrix
        Defaults to identity matrix
    n : int, optional
        Number of parameters
        Defaults to automatically inferred value
    d : int, optional
        Number of data dimensions
        Defaults to automatically inferred value
    """

    def __init__(self, *args, **kwargs):
        # Rationalise input arguments
        M = self._atleast_2d(kwargs.pop("M", None))
        m = self._atleast_1d(kwargs.pop("m", None))
        C = self._atleast_2d(kwargs.pop("C", None))
        mu = self._atleast_1d(kwargs.pop("mu", None))
        Sigma = self._atleast_2d(kwargs.pop("Sigma", None))
        n = kwargs.pop("n", 0)
        d = kwargs.pop("d", 0)

        # Determine dimensions
        n = max([n, M.shape[1], mu.shape[0], Sigma.shape[0], Sigma.shape[1]])
        d = max([d, M.shape[0], m.shape[0], C.shape[0], C.shape[1]])
        if not n:
            raise ValueError("Unable to determine number of parameters n")
        if not d:
            raise ValueError("Unable to determine data dimensions d")

        # Set defaults if no argument was passed
        M = M if M.size else np.eye(d, n)
        m = m if m.size else np.zeros(d)
        C = C if C.size else np.eye(d)
        mu = mu if mu.size else np.zeros(n)
        Sigma = Sigma if Sigma.size else np.eye(n)

        # Broadcast to correct shape
        self.M = self._broadcast_to(M, (d, n))
        self.m = np.broadcast_to(m, (d,))
        self.C = self._broadcast_to(C, (d, d))
        self.mu = np.broadcast_to(mu, (n,))
        self.Sigma = self._broadcast_to(Sigma, (n, n))

    @classmethod
    def from_joint(cls, mean, cov, n):
        """Construct model from joint distribution."""
        mean = np.atleast_1d(mean)
        cov = np.atleast_2d(cov)
        mu = mean[-n:]
        Sigma = cov[-n:, -n:]
        M = solve(Sigma, cov[-n:, :-n]).T
        m = mean[:-n] - M @ mu
        C = cov[:-n, :-n] - M @ Sigma @ M.T

        return cls(M=M, m=m, C=C, mu=mu, Sigma=Sigma)

    @property
    def n(self):
        """Dimensionality of parameter space len(theta)."""
        return self.M.shape[1]

    @property
    def d(self):
        """Dimensionality of data space len(D)."""
        return self.M.shape[0]

    def likelihood(self, theta):
        """P(D|theta) as a scipy distribution object.

        D ~ N( m + M theta, C )

        Parameters
        ----------
        theta : array_like, shape (n,)
        """
        theta = np.atleast_1d(theta)
        return multivariate_normal(self.m + self.M @ theta, self.C)

    def prior(self):
        """P(theta) as a scipy distribution object.

        theta ~ N( mu, Sigma )
        """
        return multivariate_normal(self.mu, self.Sigma)

    def posterior(self, D):
        """P(theta|D) as a scipy distribution object.

        theta ~ N( mu + Sigma M'C^{-1}(D-m), Sigma - Sigma M' C^{-1} M Sigma )

        Parameters
        ----------
        D : array_like, shape (d,)
        """
        D = np.atleast_1d(D)
        Sigma = inv(inv(self.Sigma) + self.M.T @ inv(self.C) @ self.M)
        D0 = self.m + self.M @ self.mu
        mu = self.mu + Sigma @ self.M.T @ inv(self.C) @ (D - D0)
        return multivariate_normal(mu, Sigma)

    def evidence(self):
        """P(D) as a scipy distribution object.

        D ~ N( m + M mu, C + M Sigma M' )
        """
        return multivariate_normal(
            self.m + self.M @ self.mu, self.C + self.M @ self.Sigma @ self.M.T
        )

    def joint(self):
        """P(D, theta) as a scipy distribution object.

        [  D  ] ~ N( [m + M mu]   [C + M Sigma M'  M Sigma] )
        [theta]    ( [   mu   ] , [   Sigma M'      Sigma ] )
        """
        evidence = self.evidence()
        prior = self.prior()
        mu = np.concatenate([evidence.mean, prior.mean])
        Sigma = np.block(
            [[evidence.cov, self.M @ self.Sigma], [self.Sigma @ self.M.T, prior.cov]]
        )
        return multivariate_normal(mu, Sigma)

    def DKL(self, D):
        """D_KL(P(theta|D)||P(theta)) the Kullback-Leibler divergence.

        Parameters
        ----------
        D : array_like, shape (d,)
        """
        cov_p = self.posterior(D).cov
        cov_q = self.prior().cov
        mu_p = self.posterior(D).mean
        mu_q = self.prior().mean
        return (
            -logdet(cov_p)
            + logdet(cov_q)
            + np.trace(inv(cov_q) @ cov_p - 1)
            + (mu_q - mu_p) @ inv(cov_q) @ (mu_q - mu_p)
        ) / 2

    def reduce(self, D):
        """Reduce the model to a Gaussian in the parameters.

        Parameters
        ----------
        D : array_like, shape (d,)

        Returns
        -------
        ReducedLinearModel
        """
        Sigma_L = inv(self.M.T @ inv(self.C) @ self.M)
        mu_L = Sigma_L @ self.M.T @ inv(self.C) @ (D - self.m)
        logLmax = (
            -logdet(2 * np.pi * self.C) / 2
            - (D - self.m)
            @ inv(self.C)
            @ (self.C - self.M @ Sigma_L @ self.M.T)
            @ inv(self.C)
            @ (D - self.m)
            / 2
        )
        return ReducedLinearModel(
            mu_L=mu_L,
            Sigma_L=Sigma_L,
            logLmax=logLmax,
            mu_pi=self.prior().mean,
            Sigma_pi=self.prior().cov,
        )

    def _atleast_2d(self, x):
        if x is None:
            return np.zeros(shape=(0, 0))
        return np.atleast_2d(x)

    def _atleast_1d(self, x):
        if x is None:
            return np.zeros(shape=(0,))
        return np.atleast_1d(x)

    def _broadcast_to(self, x, shape):
        if x.shape == shape:
            return x
        return x * np.eye(*shape)


class ReducedLinearModel(object):
    """A model with no data.

    If a Likelihood is Gaussian in the parameters, it is sometimes more
    clear/efficient to phrase it in terms of a parameter covariance, parameter
    mean and peak value:

    logL(theta) = logLmax - (theta - mu_L)' Sigma_L^{-1} (theta - mu_L)

    We can link this to a data-based model with the relations:

    Sigma_L = (M' C^{-1} M)^{-1}
    mu_L = Sigma_L M' C^{-1} (D-m)
    logLmax =
    - log|2 pi C|/2 - (D-m)'C^{-1}(C - M (M' C^{-1} M)^{-1} M' )C^{-1}(D-m)/2

    Parameters
    ----------
    mu_L : array_like
        Likelihood peak
    Sigma_L : array_like
        Likelihood covariance
    logLmax : float, optional
        Likelihood maximum, defaults to zero
    mu_pi : array_like, optional
        Prior mean, defaults to zero vector
    Sigma_pi : array_like, optional
        Prior covariance, defaults to identity matrix
    """

    def __init__(self, *args, **kwargs):
        self.mu_L = np.atleast_1d(kwargs.pop("mu_L"))
        self.Sigma_L = np.atleast_2d(kwargs.pop("Sigma_L", None))
        self.logLmax = kwargs.pop("logLmax", 0)
        self.mu_pi = np.atleast_1d(kwargs.pop("mu_pi", np.zeros_like(self.mu_L)))
        self.Sigma_pi = np.atleast_2d(kwargs.pop("Sigma_pi", np.eye(len(self.mu_pi))))
        self.Sigma_P = inv(inv(self.Sigma_pi) + inv(self.Sigma_L))
        self.mu_P = self.Sigma_P @ (
            solve(self.Sigma_pi, self.mu_pi) + solve(self.Sigma_L, self.mu_L)
        )

    def prior(self):
        """P(theta) as a scipy distribution object."""
        return multivariate_normal(self.mu_pi, self.Sigma_pi)

    def posterior(self):
        """P(theta|D) as a scipy distribution object."""
        return multivariate_normal(self.mu_P, self.Sigma_P)

    def logpi(self, theta):
        """P(theta) as a scalar."""
        return self.prior().logpdf(theta)

    def logP(self, theta):
        """P(theta|D) as a scalar."""
        return self.posterior().logpdf(theta)

    def logL(self, theta):
        """P(D|theta) as a scalar."""
        return (
            self.logLmax
            + multivariate_normal(self.mu_L, self.Sigma_L).logpdf(theta)
            + logdet(2 * np.pi * self.Sigma_L) / 2
        )

    def logZ(self):
        """P(D) as a scalar."""
        return (
            self.logLmax
            + logdet(self.Sigma_P) / 2
            - logdet(self.Sigma_pi) / 2
            - (self.mu_P - self.mu_pi)
            @ solve(self.Sigma_pi, self.mu_P - self.mu_pi)
            / 2
            - (self.mu_P - self.mu_L) @ solve(self.Sigma_L, self.mu_P - self.mu_L) / 2
        )

    def DKL(self):
        """D_KL(P(theta|D)||P(theta)) the Kullback-Leibler divergence."""
        return (
            logdet(self.Sigma_pi)
            - logdet(self.Sigma_P)
            + np.trace(inv(self.Sigma_pi) @ self.Sigma_P - 1)
            + (self.mu_P - self.mu_pi) @ solve(self.Sigma_pi, self.mu_P - self.mu_pi)
        ) / 2


class ReducedLinearModelUniformPrior(object):
    """A model with no data.

    Gaussian likelihood in the parameters

    logL(theta) = logLmax - (theta - mu_L)' Sigma_L^{-1} (theta - mu_L)

    Uniform prior

    We can link this to a data-based model with the relations:

    Sigma_L = (M' C^{-1} M)^{-1}
    mu_L = Sigma_L M' C^{-1} (D-m)
    logLmax =
    -log|2 pi C|/2 - (D-m)'C^{-1}(C - M (M' C^{-1} M)^{-1} M' )C^{-1}(D-m)/2

    Parameters
    ----------
    mu_L : array_like
        Likelihood peak
    Sigma_L : array_like
        Likelihood covariance
    logLmax : float, optional
        Likelihood maximum, defaults to zero
    logV : float, optional
        log prior volume, defaults to zero
    """

    def __init__(self, *args, **kwargs):
        self.mu_L = np.atleast_1d(kwargs.pop("mu_L"))
        self.Sigma_L = np.atleast_2d(kwargs.pop("Sigma_L"))
        self.logLmax = kwargs.pop("logLmax", 0)
        self.logV = kwargs.pop("logV", 0)
        self.Sigma_P = self.Sigma_L
        self.mu_P = self.mu_L

    def posterior(self):
        """P(theta|D) as a scipy distribution object."""
        return multivariate_normal(self.mu_P, self.Sigma_P)

    def logpi(self, theta):
        """P(theta) as a scalar."""
        return -self.logV

    def logP(self, theta):
        """P(theta|D) as a scalar."""
        return self.posterior().logpdf(theta)

    def logL(self, theta):
        """P(D|theta) as a scalar."""
        return (
            self.logLmax
            + logdet(2 * np.pi * self.Sigma_L) / 2
            + multivariate_normal(self.mu_L, self.Sigma_L).logpdf(theta)
        )

    def logZ(self):
        """P(D) as a scalar."""
        return self.logLmax + logdet(2 * np.pi * self.Sigma_P) / 2 - self.logV

    def DKL(self):
        """D_KL(P(theta|D)||P(theta)) the Kullback-Leibler divergence."""
        return self.logV - logdet(2 * np.pi * np.e * self.Sigma_P) / 2


class LinearMixtureModel(object):
    """A linear mixture model.

    D|theta, A ~ N( m + M theta, C )
    theta|A    ~ N( mu, Sigma )
    A          ~ categorical( exp(logA) )

    Defined by:
        Parameters:          theta (n,)
        Data:                D     (d,)
        Prior means:         mu    (k, n)
        Prior covariances:   Sigma (k, n, n)
        Data means:          m     (k, d)
        Data covariances:    C     (k, d, d)
        log mixture weights: logA  (k,)

    Parameters
    ----------
    M : array_like, optional
        if ndim==3: model matrices
        if ndim==2: model matrix with same matrix for all components
        if ndim==1: model matrix with vector diagonal for all components
        if scalar: scalar * rectangular identity matrix for all components
        Defaults to k copies of rectangular identity matrices
    m : array_like, optional
        if ndim==2: data means
        if ndim==1: data mean with same vector for all components
        if scalar: scalar * unit vector for all components
        Defaults to 0 for all components
    C : array_like, optional
        if ndim==3: data covariances
        if ndim==2: data covariance with same matrix for all components
        if ndim==1: data covariance with vector diagonal for all components
        if scalar: scalar * identity matrix for all components
        Defaults to k copies of identity matrices
    mu : array_like, optional
        if ndim==2: prior means
        if ndim==1: prior mean with same vector for all components
        if scalar: scalar * unit vector for all components
        Defaults to 0 for all components
        Prior mean, defaults to zero vector
    Sigma : array_like, optional
        if ndim==3: prior covariances
        if ndim==2: prior covariance with same matrix for all components
        if ndim==1: prior covariance with vector diagonal for all components
        if scalar: scalar * identity matrix for all components
        Defaults to k copies of identity matrices
    logA : array_like, optional
        if ndim==1: log mixture weights
        if scalar: scalar * unit vector
        Defaults to uniform weights
    n : int, optional
        Number of parameters, defaults to automatically inferred value
    d : int, optional
        Number of data dimensions, defaults to automatically inferred value
    k : int, optional
        Number of mixture components, defaults to automatically inferred value
    """

    def __init__(self, *args, **kwargs):
        # Rationalise input arguments
        M = self._atleast_3d(kwargs.pop("M", None))
        m = self._atleast_2d(kwargs.pop("m", None))
        C = self._atleast_3d(kwargs.pop("C", None))
        mu = self._atleast_2d(kwargs.pop("mu", None))
        Sigma = self._atleast_3d(kwargs.pop("Sigma", None))
        logA = self._atleast_1d(kwargs.pop("logA", None))
        n = kwargs.pop("n", 0)
        d = kwargs.pop("d", 0)
        k = kwargs.pop("k", 0)

        # Determine dimensions
        n = max([n, M.shape[2], mu.shape[1], Sigma.shape[1], Sigma.shape[2]])
        d = max([d, M.shape[1], m.shape[1], C.shape[1], C.shape[2]])
        k = max(
            [
                k,
                M.shape[0],
                m.shape[0],
                C.shape[0],
                mu.shape[0],
                Sigma.shape[0],
                logA.shape[0],
            ]
        )
        if not n:
            raise ValueError("Unable to determine number of parameters n")
        if not d:
            raise ValueError("Unable to determine data dimensions d")

        # Set defaults if no argument was passed
        M = M if M.size else np.eye(d, n)
        m = m if m.size else np.zeros(d)
        C = C if C.size else np.eye(d)
        mu = mu if mu.size else np.zeros(n)
        Sigma = Sigma if Sigma.size else np.eye(n)
        logA = logA if logA.size else -np.log(k)

        # Broadcast to correct shape
        self.M = self._broadcast_to(M, (k, d, n))
        self.m = np.broadcast_to(m, (k, d))
        self.C = self._broadcast_to(C, (k, d, d))
        self.mu = np.broadcast_to(mu, (k, n))
        self.Sigma = self._broadcast_to(Sigma, (k, n, n))
        self.logA = np.broadcast_to(logA, (k,))

    @classmethod
    def from_joint(cls, means, covs, logA, n):
        """Construct model from joint distribution."""
        mu = means[:, -n:]
        Sigma = covs[:, -n:, -n:]
        M = solve(Sigma, covs[:, -n:, :-n]).transpose(0, 2, 1)
        m = means[:, :-n] - np.einsum("ija,ia->ij", M, mu)
        C = covs[:, :-n, :-n] - np.einsum("ija,iab,ikb->ijk", M, Sigma, M)
        return cls(M=M, m=m, C=C, mu=mu, Sigma=Sigma, logA=logA)

    @property
    def n(self):
        """Dimensionality of parameter space len(theta)."""
        return self.M.shape[2]

    @property
    def d(self):
        """Dimensionality of data space len(D)."""
        return self.M.shape[1]

    @property
    def k(self):
        """Number of mixture components len(logA)."""
        return self.M.shape[0]

    def likelihood(self, theta):
        """P(D|theta) as a scipy distribution object.

        D|theta,A ~ N( m + M theta, C )
        theta|A   ~ N( mu, Sigma )
        A         ~ categorical(exp(logA))

        Parameters
        ----------
        theta : array_like, shape (n,)
        """
        theta = np.atleast_1d(theta)
        mu = self.m + np.einsum("ija,a->ij", self.M, theta)
        prior = self.prior()
        logA = prior.logpdf(theta, reduce=False) + self.logA - prior.logpdf(theta)
        return mixture_multivariate_normal(mu, self.C, logA)

    def prior(self):
        """P(theta) as a scipy distribution object.

        theta|A ~ N( mu, Sigma )
        A       ~ categorical(exp(logA))
        """
        return mixture_multivariate_normal(self.mu, self.Sigma, self.logA)

    def posterior(self, D):
        """P(theta|D) as a scipy distribution object.

        theta|D, A ~ N( mu + S M'C^{-1}(D - m - M mu), S )
        D|A        ~ N( m + M mu, C + M Sigma M' )
        A          ~ categorical(exp(logA))
        S = (Sigma^{-1} + M'C^{-1}M)^{-1}

        Parameters
        ----------
        D : array_like, shape (d,)
        """
        D = np.atleast_1d(D)
        Sigma = inv(
            inv(self.Sigma) + np.einsum("iaj,iab,ibk->ijk", self.M, inv(self.C), self.M)
        )
        D0 = self.m + np.einsum("ija,ia->ij", self.M, self.mu)
        mu = self.mu + np.einsum(
            "ija,iba,ibc,ic->ij", Sigma, self.M, inv(self.C), D - D0
        )
        evidence = self.evidence()
        logA = evidence.logpdf(D, reduce=False) + self.logA - evidence.logpdf(D)
        return mixture_multivariate_normal(mu, Sigma, logA)

    def evidence(self):
        """P(D) as a scipy distribution object.

        D|A ~ N( m + M mu, C + M Sigma M' )
        A   ~ categorical(exp(logA))
        """
        mu = self.m + np.einsum("ija,ia->ij", self.M, self.mu)
        Sigma = self.C + np.einsum("ija,iab,ikb->ijk", self.M, self.Sigma, self.M)
        return mixture_multivariate_normal(mu, Sigma, self.logA)

    def joint(self):
        """P(D, theta) as a scipy distribution object.

        [  D  ] | A ~ N( [m + M mu]   [C + M Sigma M'  M Sigma] )
        [theta] |      ( [   mu   ] , [   Sigma M'      Sigma ] )

        A           ~ categorical(exp(logA))
        """
        evidence = self.evidence()
        prior = self.prior()
        mu = np.block([evidence.means, prior.means])
        corr = np.einsum("ija,ial->ijl", self.M, self.Sigma)
        Sigma = np.block([[evidence.covs, corr], [corr.transpose(0, 2, 1), prior.covs]])
        return mixture_multivariate_normal(mu, Sigma, self.logA)

    def _atleast_3d(self, x):
        if x is None:
            return np.zeros(shape=(0, 0, 0))
        x = np.array(x)
        if x.ndim == 3:
            return x
        return np.atleast_2d(x)[None, ...]

    def _atleast_2d(self, x):
        if x is None:
            return np.zeros(shape=(0, 0))
        x = np.array(x)
        if x.ndim == 2:
            return x
        return np.atleast_1d(x)[None, ...]

    def _atleast_1d(self, x):
        if x is None:
            return np.zeros(shape=(0,))
        return np.atleast_1d(x)

    def _broadcast_to(self, x, shape):
        if x.shape == shape:
            return x
        if x.shape[1:] == shape[1:]:
            return np.broadcast_to(x, shape)
        return x * np.ones(shape) * np.eye(shape[1], shape[2])[None, ...]



import numpy as np
import pytest
import scipy.special
from numpy.testing import assert_allclose
from scipy.stats import invwishart, kstest

from lsbi.stats import mixture_multivariate_normal, multivariate_normal

N = 1000


@pytest.mark.parametrize("d", [1, 2, 5, 10])
@pytest.mark.parametrize("k", [1, 2, 5, 10])
class TestMixtureMultivariateNormal(object):
    cls = mixture_multivariate_normal

    def random(self, k, d):
        means = np.random.randn(k, d)
        covs = invwishart(scale=np.eye(d), df=d * 10).rvs(k)
        if k == 1:
            covs = np.array([covs])
        logA = np.log(scipy.stats.dirichlet(np.ones(k)).rvs())[0] + 10
        return self.cls(means, covs, logA)

    def test_rvs(self, k, d):
        dist = self.random(k, d)
        logA = dist.logA
        logA -= scipy.special.logsumexp(logA)
        mvns = [
            scipy.stats.multivariate_normal(dist.means[i], dist.covs[i])
            for i in range(k)
        ]

        samples_1, logpdfs_1 = [], []
        for _ in range(N):
            i = np.random.choice(k, p=np.exp(logA))
            x = mvns[i].rvs()
            samples_1.append(x)
            logpdf = scipy.special.logsumexp(
                [mvns[j].logpdf(x) + logA[j] for j in range(k)]
            )
            assert_allclose(logpdf, dist.logpdf(x))
            logpdfs_1.append(logpdf)
        samples_1, logpdfs_1 = np.array(samples_1), np.array(logpdfs_1)

        samples_2 = dist.rvs(N)
        logpdfs_2 = dist.logpdf(samples_2)

        for i in range(d):
            if d == 1:
                p = kstest(samples_1, samples_2).pvalue
            else:
                p = kstest(samples_1[:, i], samples_2[:, i]).pvalue
            assert p > 1e-5

        p = kstest(logpdfs_1, logpdfs_2).pvalue
        assert p > 1e-5

        for shape in [(d,), (3, d), (3, 4, d)]:
            x = np.random.rand(*shape)
            assert mvns[0].logpdf(x).shape == dist.logpdf(x).shape

    def test_bijector(self, k, d):
        dist = self.random(k, d)

        # Test inversion
        x = np.random.rand(N, d)
        theta = dist.bijector(x)
        assert_allclose(dist.bijector(theta, inverse=True), x, atol=1e-6)

        # Test sampling
        samples = dist.rvs(N)
        for i in range(d):
            if d == 1:
                p = kstest(np.squeeze(theta), samples).pvalue
            else:
                p = kstest(theta[:, i], samples[:, i]).pvalue
            assert p > 1e-5

        p = kstest(dist.logpdf(samples), dist.logpdf(theta)).pvalue
        assert p > 1e-5

        # Test shapes
        x = np.random.rand(d)
        theta = dist.bijector(x)
        assert theta.shape == x.shape
        assert dist.bijector(theta, inverse=True).shape == x.shape

        x = np.random.rand(3, 4, d)
        theta = dist.bijector(x)
        assert theta.shape == x.shape
        assert dist.bijector(theta, inverse=True).shape == x.shape

    @pytest.mark.parametrize("p", np.arange(1, 5))
    def test_marginalise_condition(self, d, k, p):
        if d <= p:
            pytest.skip("d <= p")
        i = np.random.choice(d, p, replace=False)
        j = np.array([x for x in range(d) if x not in i])
        dist = self.random(k, d)
        mixture_2 = dist.marginalise(i)
        assert mixture_2.means.shape == (k, d - p)
        assert mixture_2.covs.shape == (k, d - p, d - p)
        assert_allclose(dist.means[:, j], mixture_2.means)
        assert_allclose(dist.covs[:, j][:, :, j], mixture_2.covs)

        v = np.random.randn(k, p)
        mixture_3 = dist.condition(i, v)
        assert mixture_3.means.shape == (k, d - p)
        assert mixture_3.covs.shape == (k, d - p, d - p)

        v = np.random.randn(p)
        mixture_3 = dist.condition(i, v)
        assert mixture_3.means.shape == (k, d - p)
        assert mixture_3.covs.shape == (k, d - p, d - p)


@pytest.mark.parametrize("d", [1, 2, 5, 10])
class TestMultivariateNormal(object):
    cls = multivariate_normal

    def random(self, d):
        mean = np.random.randn(d)
        cov = invwishart(scale=np.eye(d), df=d * 10).rvs()
        return self.cls(mean, cov)

    def test_rvs(self, d):
        dist = self.random(d)
        mvn = scipy.stats.multivariate_normal(dist.mean, dist.cov)

        samples_1 = mvn.rvs(N)
        logpdfs_1 = mvn.logpdf(samples_1)
        assert_allclose(logpdfs_1, dist.logpdf(samples_1))
        samples_2 = dist.rvs(N)
        logpdfs_2 = dist.logpdf(samples_2)

        for i in range(d):
            if d == 1:
                p = kstest(samples_1, samples_2).pvalue
            else:
                p = kstest(samples_1[:, i], samples_2[:, i]).pvalue
            assert p > 1e-5

        p = kstest(logpdfs_1, logpdfs_2).pvalue
        assert p > 1e-5

        for shape in [(), (d,), (3, d), (3, 4, d)]:
            x = np.random.rand(*shape)
            assert mvn.logpdf(x).shape == dist.logpdf(x).shape

    def test_bijector(self, d):
        dist = self.random(d)
        # Test inversion
        x = np.random.rand(N, d)
        theta = dist.bijector(x)
        assert_allclose(dist.bijector(theta, inverse=True), x, atol=1e-6)

        # Test sampling
        samples = dist.rvs(N)
        for i in range(d):
            if d == 1:
                p = kstest(np.squeeze(theta), samples).pvalue
            else:
                p = kstest(theta[:, i], samples[:, i]).pvalue
            assert p > 1e-5

        p = kstest(dist.logpdf(samples), dist.logpdf(theta)).pvalue
        assert p > 1e-5

        # Test shapes
        x = np.random.rand(d)
        theta = dist.bijector(x)
        assert theta.shape == x.shape
        assert dist.bijector(theta, inverse=True).shape == x.shape

        x = np.random.rand(3, 4, d)
        theta = dist.bijector(x)
        assert theta.shape == x.shape
        assert dist.bijector(theta, inverse=True).shape == x.shape

    @pytest.mark.parametrize("p", np.arange(1, 5))
    def test_marginalise_condition_multivariate_normal(self, d, p):
        if d <= p:
            pytest.skip("d <= p")
        i = np.random.choice(d, p, replace=False)
        j = np.array([x for x in range(d) if x not in i])
        dist_1 = self.random(d)
        dist_2 = dist_1.marginalise(i)
        assert dist_2.mean.shape == (d - p,)
        assert dist_2.cov.shape == (d - p, d - p)
        assert_allclose(dist_1.mean[j], dist_2.mean)
        assert_allclose(dist_1.cov[j][:, j], dist_2.cov)

        v = np.random.randn(p)
        dist_3 = dist_1.condition(i, v)
        assert dist_3.mean.shape == (d - p,)
        assert dist_3.cov.shape == (d - p, d - p)



import numpy as np
import pytest
import torch

from lsbi.network import BinaryClassifier, BinaryClassifierBase, BinaryClassifierLPop


@pytest.mark.parametrize("input_dim", [1, 100])
@pytest.mark.parametrize("internal_dim", [16, 32])
@pytest.mark.parametrize("initial_dim", [130, 256])
class TestClassifierBase:
    CLS = BinaryClassifierBase

    @pytest.fixture
    def model(self, input_dim, internal_dim, initial_dim):
        return self.CLS(input_dim, internal_dim, initial_dim)

    @pytest.fixture
    def x(self, input_dim):
        return torch.tensor(np.random.rand(10, input_dim), dtype=torch.float32)

    @pytest.fixture
    def y(self):
        return torch.tensor(np.random.randint(0, 2, size=(10, 1)), dtype=torch.float32)

    def fit_model(self, model, input_dim):
        data_size = 10
        data = np.random.rand(data_size, input_dim)
        labels = np.random.randint(0, 2, size=(data_size))
        y_start = model.predict(data)
        model.fit(data, labels, num_epochs=1)
        y_end = model.predict(data)
        return y_start, y_end

    def test_init(self, model):
        assert isinstance(model, BinaryClassifierBase)

    def test_forward(self, model, x):
        y = model.forward(x)
        assert y.shape == (10, 1)

    def test_loss(self, model, x, y):
        with pytest.raises(NotImplementedError):
            model.loss(x, y)

    def test_predict(self, model, x):
        with pytest.raises(NotImplementedError):
            model.predict(x)

    def test_fit(self, model, x):
        with pytest.raises(NotImplementedError):
            y_start, y_end = self.fit_model(model, x.shape[1])


class TestClassifier(TestClassifierBase):
    CLS = BinaryClassifier

    def test_loss(self, model, x, y):
        loss = model.loss(x, y)
        assert loss.detach().numpy().shape == ()

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_predict(self, model, x):
        y = model.predict(x)
        assert y.shape == (10, 1)
        assert isinstance(y, np.ndarray)

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_fit(self, model, x):
        y_start, y_end = self.fit_model(model, x.shape[1])
        assert (y_start != y_end).any()


@pytest.mark.parametrize("alpha", [2, 5])
class TestClassifierLPop(TestClassifierBase):
    CLS = BinaryClassifierLPop

    @pytest.fixture
    def model(self, input_dim, internal_dim, initial_dim, alpha):
        return self.CLS(input_dim, internal_dim, initial_dim, alpha=alpha)

    def test_loss(self, model, x, y):
        loss = model.loss(x, y)
        assert loss.detach().numpy().shape == ()

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_predict(self, model, x):
        y = model.predict(x)
        assert y.shape == (10, 1)
        assert isinstance(y, np.ndarray)

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_fit(self, model, x):
        y_start, y_end = self.fit_model(model, x.shape[1])
        assert (y_start != y_end).any()



import numpy as np
import pytest
from numpy.random import rand
from numpy.testing import assert_allclose
from scipy.stats import invwishart, kstest

from lsbi.model import (
    LinearMixtureModel,
    LinearModel,
    ReducedLinearModel,
    ReducedLinearModelUniformPrior,
)

N = 1000


@pytest.mark.parametrize("n", [1, 2, 5, 10])
@pytest.mark.parametrize("d", [1, 2, 5, 10])
class TestLinearModel(object):
    cls = LinearModel

    def random(self, d, n):
        M = rand(d, n)
        m = rand(d)
        C = invwishart(scale=np.eye(d), df=10 * d).rvs()
        mu = rand(n)
        Sigma = invwishart(scale=np.eye(n), df=10 * n).rvs()
        return self.cls(M=M, m=m, C=C, mu=mu, Sigma=Sigma)

    def _test_shape(self, model, d, n):
        assert model.n == n
        assert model.d == d
        assert model.M.shape == (d, n)
        assert model.m.shape == (d,)
        assert model.C.shape == (d, d)
        assert model.mu.shape == (n,)
        assert model.Sigma.shape == (n, n)

    def test_init_M(self, d, n):
        self._test_shape(self.cls(M=rand()), 1, 1)
        self._test_shape(self.cls(M=rand(), n=n), 1, n)
        self._test_shape(self.cls(M=rand(), d=d), d, 1)
        self._test_shape(self.cls(M=rand(), d=d, n=n), d, n)
        self._test_shape(self.cls(M=rand(n)), 1, n)
        self._test_shape(self.cls(M=rand(n), d=d), d, n)
        self._test_shape(self.cls(M=rand(d, 1)), d, 1)
        self._test_shape(self.cls(M=rand(d, 1), n=n), d, n)
        self._test_shape(self.cls(M=rand(d, n)), d, n)

        M = rand()
        model = self.cls(M=M, d=d, n=n)
        assert_allclose(np.diag(model.M), M)

        M = rand(n)
        model = self.cls(M=M, d=d)
        assert_allclose(np.diag(model.M), M[: min(d, n)])

        M = rand(d, n)
        model = self.cls(M=M)
        assert_allclose(model.M, M)

    def test_init_mu(self, d, n):
        self._test_shape(self.cls(mu=rand(), d=d), d, 1)
        self._test_shape(self.cls(mu=rand(), d=d, n=n), d, n)
        self._test_shape(self.cls(mu=rand(n), d=d), d, n)

        mu = rand()
        model = self.cls(mu=mu, d=d, n=n)
        assert_allclose(model.mu, mu)

        mu = rand(n)
        model = self.cls(mu=mu, d=d)
        assert_allclose(model.mu, mu)

    def test_init_Sigma(self, d, n):
        self._test_shape(self.cls(Sigma=rand(), d=d), d, 1)
        self._test_shape(self.cls(Sigma=rand(), d=d, n=n), d, n)
        self._test_shape(self.cls(Sigma=rand(n), d=d), d, n)
        self._test_shape(self.cls(Sigma=rand(n, n), d=d), d, n)

        Sigma = rand()
        model = self.cls(Sigma=Sigma, d=d, n=n)
        assert_allclose(np.diag(model.Sigma), Sigma)

        Sigma = rand(n)
        model = self.cls(Sigma=Sigma, d=d)
        assert_allclose(np.diag(model.Sigma), Sigma)

        Sigma = rand(n, n)
        model = self.cls(Sigma=Sigma, d=d)
        assert_allclose(model.Sigma, Sigma)

    def test_init_m(self, d, n):
        self._test_shape(self.cls(m=rand(), n=n), 1, n)
        self._test_shape(self.cls(m=rand(), d=d, n=n), d, n)
        self._test_shape(self.cls(m=rand(d), n=n), d, n)

        m = rand()
        model = self.cls(m=m, d=d, n=n)
        assert_allclose(model.m, m)

        m = rand(d)
        model = self.cls(m=m, n=n)
        assert_allclose(model.m, m)

    def test_init_C(self, d, n):
        self._test_shape(self.cls(C=rand(), n=n), 1, n)
        self._test_shape(self.cls(C=rand(), d=d, n=n), d, n)
        self._test_shape(self.cls(C=rand(d), n=n), d, n)
        self._test_shape(self.cls(C=rand(d, d), n=n), d, n)

        C = rand()
        model = self.cls(C=C, d=d, n=n)
        assert_allclose(np.diag(model.C), C)

        C = rand(d)
        model = self.cls(C=C, n=n)
        assert_allclose(np.diag(model.C), C)

        C = rand(d, d)
        model = self.cls(C=C, n=n)
        assert_allclose(model.C, C)

    def test_failure(self, d, n):
        with pytest.raises(ValueError) as excinfo:
            self.cls(m=rand(d))
        string = "Unable to determine number of parameters n"
        assert string in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            self.cls(mu=rand(n))
        string = "Unable to determine data dimensions d"
        assert string in str(excinfo.value)

    def test_joint(self, d, n):
        model = self.random(d, n)
        prior = model.prior()
        evidence = model.evidence()
        joint = model.joint()

        samples_1 = prior.rvs(N)
        samples_2 = joint.rvs(N)[:, -n:]

        if n == 1:
            samples_1 = np.atleast_2d(samples_1).T

        for i in range(n):
            p = kstest(samples_1[:, i], samples_2[:, i]).pvalue
            assert p > 1e-5

        p = kstest(prior.logpdf(samples_2), prior.logpdf(samples_1)).pvalue
        assert p > 1e-5

        samples_1 = evidence.rvs(N)
        samples_2 = joint.rvs(N)[:, :d]

        if d == 1:
            samples_1 = np.atleast_2d(samples_1).T

        for i in range(d):
            p = kstest(samples_1[:, i], samples_2[:, i]).pvalue
            assert p > 1e-5

        p = kstest(evidence.logpdf(samples_2), evidence.logpdf(samples_1)).pvalue
        assert p > 1e-5

    def test_likelihood_posterior(self, d, n):
        model = self.random(d, n)
        joint = model.joint()

        samples = []
        model.prior()
        theta = model.prior().rvs()
        for _ in range(N):
            data = np.atleast_1d(model.likelihood(theta).rvs())
            theta = np.atleast_1d(model.posterior(data).rvs())
            samples.append(np.concatenate([data, theta])[:])
        samples_1 = np.array(samples)[::100]
        samples_2 = joint.rvs(len(samples_1))

        for i in range(n + d):
            p = kstest(samples_1[:, i], samples_2[:, i]).pvalue
            assert p > 1e-5

        p = kstest(joint.logpdf(samples_2), joint.logpdf(samples_1)).pvalue
        assert p > 1e-5

    def test_DKL(self, d, n):
        model = self.random(d, n)

        data = model.evidence().rvs()
        posterior = model.posterior(data)
        prior = model.prior()

        samples = posterior.rvs(N)
        Info = posterior.logpdf(samples) - prior.logpdf(samples)
        assert_allclose(Info.mean(), model.DKL(data), atol=5 * Info.std() / np.sqrt(N))

    def test_from_joint(self, d, n):
        model = self.random(d, n)
        joint = model.joint()
        mean = joint.mean
        cov = joint.cov
        model2 = self.cls.from_joint(mean, cov, n)
        assert model2.n == model.n
        assert model2.d == model.d
        assert_allclose(model2.M, model.M)
        assert_allclose(model2.m, model.m)
        assert_allclose(model2.C, model.C)
        assert_allclose(model2.mu, model.mu)
        assert_allclose(model2.Sigma, model.Sigma)

    def test_reduce(self, d, n):
        if n > d:
            pytest.skip("n > d")
        model = self.cls(M=rand(d, n))
        data = model.evidence().rvs()
        reduced_model = model.reduce(data)
        assert isinstance(reduced_model, ReducedLinearModel)
        reduced_model.prior().mean
        assert_allclose(reduced_model.prior().mean, model.prior().mean)
        assert_allclose(reduced_model.prior().cov, model.prior().cov)
        assert_allclose(reduced_model.posterior().mean, model.posterior(data).mean)
        assert_allclose(reduced_model.posterior().cov, model.posterior(data).cov)
        assert_allclose(model.evidence().logpdf(data), reduced_model.logZ())
        assert_allclose(model.DKL(data), reduced_model.DKL())

    def test_marginal_conditional(self, d, n):
        model = self.random(d, n)
        i = np.arange(d + n)[-n:]
        model_1 = model.evidence()
        model_2 = model.joint().marginalise(i)
        assert_allclose(model_1.mean, model_2.mean)
        assert_allclose(model_1.cov, model_2.cov, rtol=1e-6, atol=1e-6)

        theta = model.prior().rvs()
        model_1 = model.likelihood(theta)
        model_2 = model.joint().condition(i, theta)
        assert_allclose(model_1.mean, model_2.mean)
        assert_allclose(model_1.cov, model_2.cov, rtol=1e-6, atol=1e-6)

        i = np.arange(d + n)[:d]
        model_1 = model.prior()
        model_2 = model.joint().marginalise(i)
        assert_allclose(model_1.mean, model_2.mean)
        assert_allclose(model_1.cov, model_2.cov, rtol=1e-6, atol=1e-6)

        D = model.evidence().rvs()
        model_1 = model.posterior(D)
        model_2 = model.joint().condition(i, D)
        assert_allclose(model_1.mean, model_2.mean)
        assert_allclose(model_1.cov, model_2.cov, rtol=1e-6, atol=1e-6)

    def test_bayes_theorem(self, d, n):
        model = self.random(d, n)
        theta = model.prior().rvs()
        D = model.evidence().rvs()
        assert_allclose(
            model.posterior(D).logpdf(theta) + model.evidence().logpdf(D),
            model.likelihood(theta).logpdf(D) + model.prior().logpdf(theta),
        )


@pytest.mark.parametrize("n", np.arange(1, 6))
class TestReducedLinearModel(object):
    def random(self, n):
        mu_pi = np.random.randn(n)
        Sigma_pi = invwishart(scale=np.eye(n)).rvs()
        mu_L = np.random.randn(n)
        Sigma_L = invwishart(scale=np.eye(n)).rvs()
        logLmax = np.random.randn()

        return ReducedLinearModel(
            mu_pi=mu_pi, Sigma_pi=Sigma_pi, logLmax=logLmax, mu_L=mu_L, Sigma_L=Sigma_L
        )

    def test_bayes_theorem(self, n):
        model = self.random(n)
        theta = model.prior().rvs()
        assert_allclose(
            model.logP(theta) + model.logZ(), model.logL(theta) + model.logpi(theta)
        )


@pytest.mark.parametrize("n", np.arange(1, 6))
class TestReducedLinearModelUniformPrior(object):
    def random(self, n):
        mu_L = np.random.randn(n)
        Sigma_L = invwishart(scale=np.eye(n)).rvs()
        logLmax = np.random.randn()
        logV = np.random.randn()

        return ReducedLinearModelUniformPrior(
            logLmax=logLmax, logV=logV, mu_L=mu_L, Sigma_L=Sigma_L
        )

    def test_model(self, n):
        model = self.random(n)
        theta = model.posterior().rvs(N)
        assert_allclose(
            model.logpi(theta) + model.logL(theta), model.logP(theta) + model.logZ()
        )

        logV = 50
        Sigma_pi = np.exp(2 * logV / n) / (2 * np.pi) * np.eye(n)

        reduced_model = ReducedLinearModel(
            logLmax=model.logLmax,
            mu_L=model.mu_L,
            Sigma_L=model.Sigma_L,
            Sigma_pi=Sigma_pi,
        )

        model = ReducedLinearModelUniformPrior(
            logLmax=model.logLmax, mu_L=model.mu_L, Sigma_L=model.Sigma_L, logV=logV
        )

        assert_allclose(reduced_model.logZ(), model.logZ())
        assert_allclose(reduced_model.DKL(), model.DKL())

    def test_bayes_theorem(self, n):
        model = self.random(n)
        theta = model.posterior().rvs()
        assert_allclose(
            model.logP(theta) + model.logZ(), model.logL(theta) + model.logpi(theta)
        )


@pytest.mark.parametrize("k", np.arange(1, 6))
@pytest.mark.parametrize("d", np.arange(1, 6))
@pytest.mark.parametrize("n", np.arange(1, 6))
class TestLinearMixtureModel(object):
    cls = LinearMixtureModel

    def random(self, k, d, n):
        M = rand(k, d, n)
        m = rand(k, d)
        C = np.array(
            [
                np.atleast_2d(invwishart(scale=np.eye(d), df=d * 10).rvs())
                for _ in range(k)
            ]
        )

        mu = rand(k, n)
        Sigma = np.array(
            [
                np.atleast_2d(invwishart(scale=np.eye(n), df=d * 10).rvs())
                for _ in range(k)
            ]
        )
        logA = np.log(rand(k))
        return self.cls(M=M, m=m, C=C, mu=mu, Sigma=Sigma, logA=logA)

    def _test_shape(self, model, k, d, n):
        assert model.n == n
        assert model.d == d
        assert model.k == k
        assert model.M.shape == (k, d, n)
        assert model.m.shape == (
            k,
            d,
        )
        assert model.C.shape == (k, d, d)
        assert model.mu.shape == (
            k,
            n,
        )
        assert model.Sigma.shape == (k, n, n)
        assert model.logA.shape == (k,)

    def test_init_M(self, k, d, n):
        self._test_shape(self.cls(M=rand()), 1, 1, 1)
        self._test_shape(self.cls(M=rand(), n=n), 1, 1, n)
        self._test_shape(self.cls(M=rand(), d=d), 1, d, 1)
        self._test_shape(self.cls(M=rand(), k=k), k, 1, 1)
        self._test_shape(self.cls(M=rand(), d=d, n=n), 1, d, n)
        self._test_shape(self.cls(M=rand(), k=k, n=n), k, 1, n)
        self._test_shape(self.cls(M=rand(), k=k, d=d), k, d, 1)
        self._test_shape(self.cls(M=rand(), k=k, d=d, n=n), k, d, n)
        self._test_shape(self.cls(M=rand(n)), 1, 1, n)
        self._test_shape(self.cls(M=rand(n), d=d), 1, d, n)
        self._test_shape(self.cls(M=rand(n), k=k), k, 1, n)
        self._test_shape(self.cls(M=rand(n), k=k, d=d), k, d, n)

        self._test_shape(self.cls(M=rand(d, 1)), 1, d, 1)
        self._test_shape(self.cls(M=rand(d, 1), k=k), k, d, 1)
        self._test_shape(self.cls(M=rand(d, 1), n=n), 1, d, n)
        self._test_shape(self.cls(M=rand(d, 1), k=k, n=n), k, d, n)

        self._test_shape(self.cls(M=rand(k, 1, 1)), k, 1, 1)
        self._test_shape(self.cls(M=rand(k, 1, 1), d=d), k, d, 1)
        self._test_shape(self.cls(M=rand(k, 1, 1), n=n), k, 1, n)
        self._test_shape(self.cls(M=rand(k, 1, 1), d=d, n=n), k, d, n)

        self._test_shape(self.cls(M=rand(k, d, 1)), k, d, 1)
        self._test_shape(self.cls(M=rand(k, d, 1), n=n), k, d, n)

        self._test_shape(self.cls(M=rand(k, 1, n)), k, 1, n)
        self._test_shape(self.cls(M=rand(k, 1, n), d=d), k, d, n)

        self._test_shape(self.cls(M=rand(1, d, n)), 1, d, n)
        self._test_shape(self.cls(M=rand(1, d, n), k=k), k, d, n)

        self._test_shape(self.cls(M=rand(d, n)), 1, d, n)
        self._test_shape(self.cls(M=rand(d, n), k=k), k, d, n)

        self._test_shape(self.cls(M=rand(k, d, n)), k, d, n)

        M = rand()
        model = self.cls(M=M, d=d, n=n)
        assert_allclose(np.diag(model.M[0]), M)

        M = rand(n)
        model = self.cls(M=M, d=d)
        assert_allclose(np.diag(model.M[0]), M[: min(d, n)])

        M = rand(d, n)
        model = self.cls(M=M)
        assert_allclose(model.M[0], M)

        M = rand()
        model = self.cls(M=M, k=k, d=d, n=n)
        for M_ in model.M:
            assert_allclose(M_, model.M[0])
            assert_allclose(np.diag(M_), M)

        M = rand(n)
        model = self.cls(M=M, d=d)
        for M_ in model.M:
            assert_allclose(M_, model.M[0])
            assert_allclose(np.diag(M_), M[: min(d, n)])

        M = rand(d, n)
        model = self.cls(M=M)
        for M_ in model.M:
            assert_allclose(M_, model.M[0])
            assert_allclose(M_, M)

        M = rand(k, d, n)
        model = self.cls(M=M)
        assert_allclose(model.M, M)

    def test_init_mu(self, k, d, n):
        self._test_shape(self.cls(mu=rand(), d=d), 1, d, 1)
        self._test_shape(self.cls(mu=rand(), k=k, d=d), k, d, 1)
        self._test_shape(self.cls(mu=rand(), d=d, n=n), 1, d, n)
        self._test_shape(self.cls(mu=rand(), k=k, d=d, n=n), k, d, n)
        self._test_shape(self.cls(mu=rand(n), d=d), 1, d, n)
        self._test_shape(self.cls(mu=rand(n), k=k, d=d), k, d, n)
        self._test_shape(self.cls(mu=rand(k, n), d=d), k, d, n)

        mu = rand()
        model = self.cls(mu=mu, d=d, n=n)
        assert_allclose(model.mu, mu)

        mu = rand(n)
        model = self.cls(mu=mu, d=d)
        assert_allclose(model.mu[0], mu)

        mu = rand()
        model = self.cls(mu=mu, k=k, d=d, n=n)
        assert_allclose(model.mu, mu)

        mu = rand(n)
        model = self.cls(mu=mu, k=k, d=d)
        for mu_ in model.mu:
            assert_allclose(mu_, mu)

        mu = rand(k, n)
        model = self.cls(mu=mu, d=d)
        assert_allclose(model.mu, mu)

    def test_init_Sigma(self, k, d, n):
        self._test_shape(self.cls(Sigma=rand(), d=d), 1, d, 1)
        self._test_shape(self.cls(Sigma=rand(), k=k, d=d), k, d, 1)
        self._test_shape(self.cls(Sigma=rand(), d=d, n=n), 1, d, n)
        self._test_shape(self.cls(Sigma=rand(), k=k, d=d, n=n), k, d, n)
        self._test_shape(self.cls(Sigma=rand(n), d=d), 1, d, n)
        self._test_shape(self.cls(Sigma=rand(n), k=k, d=d), k, d, n)
        self._test_shape(self.cls(Sigma=rand(n, n), d=d), 1, d, n)
        self._test_shape(self.cls(Sigma=rand(n, n), k=k, d=d), k, d, n)
        self._test_shape(self.cls(Sigma=rand(k, n, n), d=d), k, d, n)

        Sigma = rand()
        model = self.cls(Sigma=Sigma, d=d, n=n)
        assert_allclose(np.diag(model.Sigma[0]), Sigma)

        Sigma = rand(n)
        model = self.cls(Sigma=Sigma, d=d)
        assert_allclose(np.diag(model.Sigma[0]), Sigma)

        Sigma = rand(n, n)
        model = self.cls(Sigma=Sigma, d=d)
        assert_allclose(model.Sigma[0], Sigma)

        Sigma = rand()
        model = self.cls(Sigma=Sigma, k=k, d=d, n=n)
        for Sigma_ in model.Sigma:
            assert_allclose(np.diag(Sigma_), Sigma)

        Sigma = rand(n)
        model = self.cls(Sigma=Sigma, k=k, d=d)
        for Sigma_ in model.Sigma:
            assert_allclose(np.diag(Sigma_), Sigma)

        Sigma = rand(n, n)
        model = self.cls(Sigma=Sigma, k=k, d=d)
        for Sigma_ in model.Sigma:
            assert_allclose(Sigma_, Sigma)

        Sigma = rand(k, n, n)
        model = self.cls(Sigma=Sigma, d=d)
        assert_allclose(model.Sigma, Sigma)

    def test_init_m(self, k, d, n):
        self._test_shape(self.cls(m=rand(), n=n), 1, 1, n)
        self._test_shape(self.cls(m=rand(), k=k, n=n), k, 1, n)
        self._test_shape(self.cls(m=rand(), n=n, d=d), 1, d, n)
        self._test_shape(self.cls(m=rand(), k=k, n=n, d=d), k, d, n)
        self._test_shape(self.cls(m=rand(d), n=n), 1, d, n)
        self._test_shape(self.cls(m=rand(d), k=k, n=n), k, d, n)
        self._test_shape(self.cls(m=rand(k, d), n=n), k, d, n)

        m = rand()
        model = self.cls(m=m, d=d, n=n)
        assert_allclose(model.m, m)

        m = rand(d)
        model = self.cls(m=m, n=n)
        assert_allclose(model.m[0], m)

        m = rand()
        model = self.cls(m=m, k=k, d=d, n=n)
        assert_allclose(model.m, m)

        m = rand(d)
        model = self.cls(m=m, k=k, n=n)
        for mu_ in model.m:
            assert_allclose(mu_, m)

        m = rand(k, d)
        model = self.cls(m=m, n=n)
        assert_allclose(model.m, m)

    def test_init_C(self, k, d, n):
        self._test_shape(self.cls(C=rand(), n=n), 1, 1, n)
        self._test_shape(self.cls(C=rand(), k=k, n=n), k, 1, n)
        self._test_shape(self.cls(C=rand(), n=n, d=d), 1, d, n)
        self._test_shape(self.cls(C=rand(), k=k, n=n, d=d), k, d, n)
        self._test_shape(self.cls(C=rand(d), n=n), 1, d, n)
        self._test_shape(self.cls(C=rand(d), k=k, n=n), k, d, n)
        self._test_shape(self.cls(C=rand(d, d), n=n), 1, d, n)
        self._test_shape(self.cls(C=rand(d, d), k=k, n=n), k, d, n)
        self._test_shape(self.cls(C=rand(k, d, d), n=n), k, d, n)

        C = rand()
        model = self.cls(C=C, d=d, n=n)
        assert_allclose(np.diag(model.C[0]), C)

        C = rand(d)
        model = self.cls(C=C, n=n)
        assert_allclose(np.diag(model.C[0]), C)

        C = rand(d, d)
        model = self.cls(C=C, n=n)
        assert_allclose(model.C[0], C)

        C = rand()
        model = self.cls(C=C, k=k, d=d, n=n)
        for Sigma_ in model.C:
            assert_allclose(np.diag(Sigma_), C)

        C = rand(d)
        model = self.cls(C=C, k=k, n=n)
        for Sigma_ in model.C:
            assert_allclose(np.diag(Sigma_), C)

        C = rand(d, d)
        model = self.cls(C=C, k=k, n=n)
        for Sigma_ in model.C:
            assert_allclose(Sigma_, C)

        C = rand(k, d, d)
        model = self.cls(C=C, n=n)
        assert_allclose(model.C, C)

    def test_init_logA(self, k, d, n):
        self._test_shape(self.cls(logA=rand(), d=d, n=n), 1, d, n)
        self._test_shape(self.cls(logA=rand(), k=k, d=d, n=n), k, d, n)
        self._test_shape(self.cls(logA=rand(k), d=d, n=n), k, d, n)

        logA = rand()
        model = self.cls(logA=logA, d=d, n=n)
        assert_allclose(model.logA, logA)

        logA = rand(k)
        model = self.cls(logA=logA, d=d, n=n)
        assert_allclose(model.logA, logA)

    def test_failure(self, k, d, n):
        with pytest.raises(ValueError) as excinfo:
            self.cls(m=rand(d))
        string = "Unable to determine number of parameters n"
        assert string in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            self.cls(mu=rand(n))
        string = "Unable to determine data dimensions d"
        assert string in str(excinfo.value)

    def test_joint(self, k, d, n):
        model = self.random(k, d, n)
        prior = model.prior()
        evidence = model.evidence()
        joint = model.joint()

        samples_1 = prior.rvs(N)
        samples_2 = joint.rvs(N)[:, -n:]

        if n == 1:
            samples_1 = np.atleast_2d(samples_1).T

        for i in range(n):
            p = kstest(samples_1[:, i], samples_2[:, i]).pvalue
            assert p > 1e-5

        p = kstest(prior.logpdf(samples_2), prior.logpdf(samples_1)).pvalue
        assert p > 1e-5

        samples_1 = evidence.rvs(N)
        samples_2 = joint.rvs(N)[:, :d]

        if d == 1:
            samples_1 = np.atleast_2d(samples_1).T

        for i in range(d):
            p = kstest(samples_1[:, i], samples_2[:, i]).pvalue
            assert p > 1e-5

        p = kstest(evidence.logpdf(samples_2), evidence.logpdf(samples_1)).pvalue
        assert p > 1e-5

    def test_likelihood_posterior(self, k, d, n):
        model = self.random(k, d, n)
        joint = model.joint()

        samples = []
        model.prior()
        theta = model.prior().rvs()
        for _ in range(N):
            data = np.atleast_1d(model.likelihood(theta).rvs())
            theta = np.atleast_1d(model.posterior(data).rvs())
            samples.append(np.concatenate([data, theta])[:])
        samples_1 = np.array(samples)[::100]
        samples_2 = joint.rvs(len(samples_1))

        for i in range(n + d):
            p = kstest(samples_1[:, i], samples_2[:, i]).pvalue
            assert p > 1e-5

        p = kstest(joint.logpdf(samples_2), joint.logpdf(samples_1)).pvalue
        assert p > 1e-5

    def test_from_joint(self, k, d, n):
        model = self.random(k, d, n)
        joint = model.joint()
        means = joint.means
        covs = joint.covs
        logA = joint.logA
        model2 = self.cls.from_joint(means, covs, logA, n)
        assert model2.n == model.n
        assert model2.d == model.d
        assert_allclose(model2.M, model.M)
        assert_allclose(model2.m, model.m)
        assert_allclose(model2.C, model.C)
        assert_allclose(model2.mu, model.mu)
        assert_allclose(model2.Sigma, model.Sigma)
        assert_allclose(model2.logA, model.logA)

    def test_marginal_conditional(self, k, d, n):
        model = self.random(k, d, n)
        i = np.arange(d + n)[-n:]
        model_1 = model.evidence()
        model_2 = model.joint().marginalise(i)
        assert_allclose(model_1.means, model_2.means)
        assert_allclose(model_1.covs, model_2.covs, rtol=1e-6, atol=1e-6)
        assert_allclose(model_1.logA, model_2.logA)

        theta = model.prior().rvs()
        model_1 = model.likelihood(theta)
        model_2 = model.joint().condition(i, theta)
        assert_allclose(model_1.means, model_2.means)
        assert_allclose(model_1.covs, model_2.covs, rtol=1e-6, atol=1e-6)
        assert_allclose(model_1.logA, model_2.logA)

        i = np.arange(d + n)[:d]
        model_1 = model.prior()
        model_2 = model.joint().marginalise(i)
        assert_allclose(model_1.means, model_2.means)
        assert_allclose(model_1.covs, model_2.covs, rtol=1e-6, atol=1e-6)
        assert_allclose(model_1.logA, model_2.logA)

        D = model.evidence().rvs()
        model_1 = model.posterior(D)
        model_2 = model.joint().condition(i, D)
        assert_allclose(model_1.means, model_2.means)
        assert_allclose(model_1.covs, model_2.covs, rtol=1e-6, atol=1e-6)
        assert_allclose(model_1.logA, model_2.logA)

    def test_bayes_theorem(self, k, d, n):
        model = self.random(k, d, n)
        theta = model.prior().rvs()
        D = model.evidence().rvs()
        assert_allclose(
            model.posterior(D).logpdf(theta) + model.evidence().logpdf(D),
            model.likelihood(theta).logpdf(D) + model.prior().logpdf(theta),
        )



"""Utility functions for lsbi."""
import numpy as np


def logdet(A):
    """log(abs(det(A)))."""
    return np.linalg.slogdet(A)[1]


def quantise(f, x, tol=1e-8):
    """Quantise f(x) to zero within tolerance tol."""
    y = np.atleast_1d(f(x))
    return np.where(np.abs(y) < tol, 0, y)


def bisect(f, a, b, args=(), tol=1e-8):
    """Vectorised simple bisection search.

    The shape of the output is the broadcasted shape of a and b.

    Parameters
    ----------
    f : callable
        Function to find the root of.
    a : array_like
        Lower bound of the search interval.
    b : array_like
        Upper bound of the search interval.
    args : tuple, optional
        Extra arguments to `f`.
    tol : float, optional
        (absolute) tolerance of the solution

    Returns
    -------
    x : ndarray
        Solution to the equation f(x) = 0.
    """
    a = np.array(a)
    b = np.array(b)
    while np.abs(a - b).max() > tol:
        fa = quantise(f, a, tol)
        fb = quantise(f, b, tol)
        a = np.where(fb == 0, b, a)
        b = np.where(fa == 0, a, b)

        if np.any(fa * fb > 0):
            raise ValueError("f(a) and f(b) must have opposite signs")
        q = (a + b) / 2
        fq = quantise(f, q, tol)

        a = np.where(fq == 0, q, a)
        a = np.where(fa * fq > 0, q, a)

        b = np.where(fq == 0, q, b)
        b = np.where(fb * fq > 0, q, b)
    return (a + b) / 2
